STATRACK = {}

STATRACK.DEBUG = false

local StatRackFrame = CreateFrame("FRAME", "STATRACK", UIParent)
local frame = StatRackFrame

local TUBE_RACK_MIN_X, TUBE_RACK_MAX_X = 65, 256
local TUBE_RACK_MIN_Y, TUBE_RACK_MAX_Y = 65, 256

local isLoaded = false

-- The following needs a lot of refactoring, readability and modularity, most things are tweaked by hand, while I've learned the graphical side of WoW API.

-- statTubes
local statTubesData
local TUBES_CRIT, TUBES_HASTE, TUBES_MAST, TUBES_VERS, TUBES_PRIM = 1, 2, 3, 4, 5
local TUBES_NAME_IX, TUBES_COLOR_IX, TUBES_X_OFFSET_IX, TUBES_LETTER_IX = 1, 2, 3, 4
local thisBubbleOverlayNumber = 1
statTubesData = {
	{
		"HASTE",
		{0.18, 0.45, 0.18, 0.95},
		-0.328,
		"H"
	},
	{
		"CRIT",
		{0.75, 0.05, 0.05, 0.95},
		-0.159,
		"C"
	},
	{
		"MAST",
		{0.33, 0.1, 0.33, 0.95},
		0.009,
		"M"
	},
	{
		"VERS",
		{0.28, 0.72, 0.33, 0.95},
		0.1765,
		"V"
	},
	{
		"PRIM",
		{0.05, 0.1, 0.75, 0.95},
		0.3395,
		"P"
	}
}

-- SD Indexes
STATRACK.statsDisplay = nil
local SD_ID, SD_FUNC, SD_VAL, SD_COLOR, SD_TUBE_FRAME = 1, 2, 3, 4, 5
-- Find primary stat index
-- Initialize statsDisplay 
-- (The single next iteration (1, 2, .., 5) is computed when OnUpdate frameskip ends, to spread load)
STATRACK.statsDisplay = {
	{
		"Haste", 
		function() return UnitSpellHaste("player") end, 
		nil,
		"808080",
		hasteTube
	},
	{
		"Crit", 
		function() return GetCritChance() end, 
		nil,
		"808080",
		critTube
	}, 
	{
		"Mastery", 
		function() return select(1, GetMasteryEffect()) end, 
		nil,
		"808080",
		masteryTube
	},
	{
		"Vers", 
		function() return GetCombatRating(29) end, 
		nil,
		"808080",
		versatilityTube
	},
	{
		"Primary", 
		function() return UnitStat("player", STATRACK.statsDisplay.statIndex) end, 
		nil,
		"808080",
		primaryTube
	}
}
STATRACK.statsDisplay.tableSize = 25
STATRACK.statsDisplay.statIndex = mainStatIndex

local statsDisplay = STATRACK.statsDisplay

local NUM_TUBE_FRAMES = 5
local NUM_BUBBLE_TEX = 4

local tubeRack
local hasteTube
local critTube
local masteryTube
local versatilityTube
local primaryTube
local hideButton
local lockButton
local welcome

frame.limiter = 0.1
frame.nToCompute = 1
function StatRackFrameTicker(self, elapsed)
    frame.limiter = frame.limiter - elapsed
	if frame.limiter > 0 then return end
	frame.limiter = InCombatLockdown() and 0.05 or 0.1
	
	local sd = statsDisplay[frame.nToCompute]
	local thisTube = sd[SD_TUBE_FRAME]
	
	local currValue = sd[SD_FUNC]()
	
	local thisRating = STATRACK.ValueOfIntAbscond(
		sd[SD_ID], 
		currValue, 
		statsDisplay.tableSize
	)
	sd[SD_VAL] = currValue
	
	local letterFrame = thisTube.letterFrame
	local fontString = letterFrame.fontString
	if thisRating > 0.75 then
		if letterFrame.mode ~= "FAST" then
			fontString:SetTextColor(1.0, 0.1, 0.1, 1.0)
			fontString.animGroup.alpha:SetDuration(0.35)
			fontString:Show()
			fontString.animGroup:Play()
			letterFrame.mode = "FAST"
		end
	elseif thisRating > 0.45 then
		if letterFrame.mode ~= "SLOW" then
			fontString:SetTextColor(0.9, 0.5, 0.5, 1.0)
			fontString.animGroup.alpha:SetDuration(0.9)
			fontString:Show()
			fontString.animGroup:Play()
			letterFrame.mode = "SLOW"
		end
	elseif fontString.animGroup:IsPlaying() then
		fontString.animGroup:Stop()
		fontString:Hide()
		letterFrame.mode = "OFF"
	end
	
	if (displayIsNumerical) then
		local red = thisRating
		local green = 1.0 - thisRating
		local blue = green
		sd[SD_COLOR] = string.format("%s%s%s", red, green, blue)
	end
	
	frame.nToCompute = max(1, (frame.nToCompute+1) % (#statsDisplay+1))
	
	SetNewScaleAnim(thisTube, thisRating) 
	
	-- Bubble stuff
	thisTube.bubbleTextures[thisBubbleOverlayNumber]:Show()
	for i=1,NUM_BUBBLE_TEX-1,1 do
		thisBubbleOverlayNumber = (thisBubbleOverlayNumber % NUM_BUBBLE_TEX) + 1
		thisTube.bubbleTextures[thisBubbleOverlayNumber]:Hide()
	end
	thisBubbleOverlayNumber = ( ( thisBubbleOverlayNumber + 1 ) % NUM_BUBBLE_TEX) + 1 -- additional increment will correctly sequence 1-4, because we jump 4+1 times.
end

local function StartMoving(...)
	frame.StartMoving(...)
	--[[for i=1,NUM_TUBE_FRAMES,1 do
		statsDisplay[i][SD_TUBE_FRAME].StartMoving(...)
	end]]
	--resizeButton.StartMoving(...)
end

local function StopMovingOrSizing(...)
	frame.StopMovingOrSizing(...)
	for i=1,NUM_TUBE_FRAMES,1 do
		statsDisplay[i][SD_TUBE_FRAME].StopMovingOrSizing(...)
	end
	resizeButton.StopMovingOrSizing(...)
end

function STATRACK.Load()
	frame:ClearAllPoints()
	do
		local userPos = STATRACK.GetVar("userPos", true)
		if userPos ~= nil and #userPos == 4 then
			local x1, y1, x2, y2 = unpack(userPos)
			frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", x1, y1)
			frame:SetWidth(x2)
			frame:SetHeight(y2)
		else
			frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", UIParent:GetWidth()*0.33, UIParent:GetHeight()*0.5)
			frame:SetWidth(128)
			frame:SetHeight(110)
		end
	end
	frame:SetClampedToScreen(true)
	if tubeRack == nil then
		tubeRack = CreateFrame("FRAME", "STATRACK_TUBE_RACK", frame)
	end
	tubeRack:SetAllPoints(frame)
	tubeRack:SetFrameStrata("BACKGROUND")
	
	if tubeRack.texture == nil then 
		tubeRack.texture = tubeRack:CreateTexture("STATRACK_TUBE_RACK_TEX", "BACKGROUND")
	end
	tubeRack.texture:SetAllPoints(tubeRack)
	tubeRack.texture:SetTexture("Interface/AddOns/StatRack/Textures/tube_rack.blp")

	tubeRack.tubeWidthScaleX = 0.07
	tubeRack.tubeHeightScaleY = 0.634
	tubeRack.tubeOffsetY = 0.142

	-- Initialize Tubes
	for q=1,NUM_TUBE_FRAMES,1 do
		local tubeName, tubeColor, tubeXOffset, tubeLetter = unpack(statTubesData[q])
		if statsDisplay[q][SD_TUBE_FRAME] == nil then
			statsDisplay[q][SD_TUBE_FRAME] = CreateFrame("FRAME", string.format("STATRACK_%s", tubeName), frame)
		end
		local tf = statsDisplay[q][SD_TUBE_FRAME]
		
		tf:ClearAllPoints()
		tf:SetFrameStrata("LOW")
		tf:SetWidth(frame:GetWidth()*tubeRack.tubeWidthScaleX)
		tf:SetHeight(frame:GetHeight()*tubeRack.tubeHeightScaleY)
		tf:SetPoint("CENTER", frame, "CENTER", frame:GetWidth()*tubeXOffset, 0)
		tf:SetPoint("BOTTOM", frame, "BOTTOM", 0, frame:GetHeight()*tubeRack.tubeOffsetY)
		if tf.texture == nil then 
			tf.texture = tf:CreateTexture(string.format("STATRACK_%s_TEX", tubeName), "ARTWORK")
		end
		tf.texture:SetAllPoints(tf)
		tf.texture:SetColorTexture(unpack(tubeColor))
		tf.bubbleTextures = {}
		for r=1,NUM_BUBBLE_TEX,1 do
			if tf.bubbleTextures[r] == nil then 
				tf.bubbleTextures[r] = tf:CreateTexture(string.format("STATRACK_%s_BUBBLE%d_TEX", tubeName, r), "OVERLAY")
				tf.bubbleTextures[r]:SetTexture(string.format("Interface/AddOns/StatRack/Textures/tube_bubble_overlay%d.blp", r), "REPEAT", "REPEAT")
			end
			tf.bubbleTextures[r]:SetPoint("BOTTOMLEFT", tf, "BOTTOMLEFT")
			tf.bubbleTextures[r]:SetPoint("TOPRIGHT", tf, "TOPRIGHT")
			tf.bubbleTextures[r]:SetVertTile(true)
			tf.bubbleTextures[r]:SetAlpha(0.3)
		end
		if tf.animGroup == nil then 
			tf.animGroup = tf:CreateAnimationGroup(string.format("STATRACK_%s_ANIM_GROUP", tubeName))
		end
		if tf.animGroup.scale == nil then 
			tf.animGroup.scale = tf.animGroup:CreateAnimation("SCALE", string.format("STATRACK_%s_SCALE", tubeName))
		end
		tf.animGroup.scale:SetDuration(0.25)
		tf.animGroup.scale:SetSmoothing("IN_OUT")
		if tf.animGroup.translation == nil then 
			tf.animGroup.translation = tf.animGroup:CreateAnimation("TRANSLATION", string.format("STATRACK_%s_TRANSLATION", tubeName))
		end
		tf.animGroup.translation:SetDuration(0.25)
		tf.animGroup.translation:SetSmoothing("IN_OUT")
		tf.scale = 1.0
		
		if tf.letterFrame == nil then
			tf.letterFrame = CreateFrame("FRAME", string.format("STATRACK_%s_LETTER", tubeName))
		end
		if tf.letterFrame.fontString == nil then 
			tf.letterFrame.fontString = tf.letterFrame:CreateFontString(string.format("STATRACK_%s_LETTER_FS", tubeName), "ARTWORK")
		end
		tf.letterFrame.fontString:SetFontObject("GameFontNormal")
		tf.letterFrame.fontString:SetFont("Interface/AddOns/StatRack/Fonts/Langdon.otf", 16)
		tf.letterFrame.fontString:SetTextHeight(tf:GetWidth()*1.1906)
		tf.letterFrame.fontString:SetPoint("BOTTOM", tf, "BOTTOM", 0, frame:GetHeight()*0.52)
		tf.letterFrame.fontString:SetText(tubeLetter)
		tf.letterFrame.fontString:SetTextColor(0.7, 0.4, 0.4, 1.0)
		if tf.letterFrame.fontString.animGroup == nil then 
			tf.letterFrame.fontString.animGroup = tf.letterFrame.fontString:CreateAnimationGroup(string.format("STATRACK_%s_LETTER_ANIM_GROUP", tubeName))
		end
		if tf.letterFrame.fontString.animGroup.alpha == nil then 
			tf.letterFrame.fontString.animGroup.alpha = tf.letterFrame.fontString.animGroup:CreateAnimation("ALPHA", string.format("STATRACK_%s_LETTER_ALPHA", tubeName))
		end
		tf.letterFrame.fontString.animGroup.alpha:SetDuration(0.9)
		tf.letterFrame.fontString.animGroup.alpha:SetFromAlpha(0.2)
		tf.letterFrame.fontString.animGroup.alpha:SetToAlpha(1.0)
		tf.letterFrame.fontString.animGroup.alpha:SetSmoothing("IN_OUT")
		tf.letterFrame.fading = false
		tf.letterFrame.mode = "OFF"
		tf.letterFrame.fontString.animGroup:SetScript("OnFinished",
			function()
				if tf.letterFrame.fading then
					--tf.letterFrame:SetAlpha(0.2)
					tf.letterFrame.fontString.animGroup.alpha:SetFromAlpha(0.2)
					tf.letterFrame.fontString.animGroup.alpha:SetToAlpha(1.0)
					tf.letterFrame.fading = false
					tf.letterFrame.fontString.animGroup:Play()
				else
					--tf.letterFrame:SetAlpha(1.0)
					tf.letterFrame.fontString.animGroup.alpha:SetFromAlpha(1.0)
					tf.letterFrame.fontString.animGroup.alpha:SetToAlpha(0.2)
					tf.letterFrame.fading = true
					tf.letterFrame.fontString.animGroup:Play()
				end
			end
		)
		tf.letterFrame.fontString.animGroup:Play()
	end

	-- Initialize utility buttons
	if resizeButton == nil then
		resizeButton = CreateFrame("BUTTON", "STATRACK_RESIZE_BUTTON", frame)
	end
	resizeButton:SetFrameStrata("HIGH")
	resizeButton:SetWidth(12)
	resizeButton:SetHeight(12)
	resizeButton:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT")
	resizeButton:SetNormalTexture("Interface/AddOns/StatRack/Textures/resize_button.blp")
	resizeButton:SetAlpha(0.5)
	resizeButton:RegisterForClicks("LeftButtonUp", "LeftButtonDown")
	resizeButton:RegisterForDrag("LeftButton")
	resizeButton:SetMovable(true)
	resizeButton:EnableMouse(true)
	resizeButton:SetScript("OnClick", 
		function(self, button, down)
			local StopHighlight = function()
				resizeButton:SetAlpha(0.5)
				resizeButton:ClearAllPoints()
				resizeButton:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT")
				resizeButton:SetScript("OnMouseUp", nil)
				resizeButton:SetScript("OnDragStart", nil)
				resizeButton:StopMovingOrSizing()
				resizeButton:SetScript("OnDragStop", nil)
				resizeButton:SetButtonState("NORMAL", false)
			end
			
			if button == "LeftButton" and down then
				resizeButton:SetButtonState("PUSHED", true)
				
				resizeButton:SetAlpha(0.9)
				
				resizeButton:SetScript("OnDragStart",
					function()
						resizeButton:ClearAllPoints()
						resizeButton:StartMoving()
					end
				)
				resizeButton:SetScript("OnDragStop",
					function()
						resizeButton:ClearAllPoints()
						resizeButton:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT")
						resizeButton:StopMovingOrSizing()
						StopHighlight()
					end
				)
				
				resizeButton.limit = 0.1
				resizeButton.x1 = frame:GetLeft()
				resizeButton.x2 = frame:GetRight()
				resizeButton.y1 = frame:GetBottom()
				resizeButton.y2 = frame:GetTop()
				resizeButton:SetScript("OnUpdate", 
					function(self, elapsed)
						resizeButton.limit = resizeButton.limit - elapsed
						if resizeButton.limit > 0 then return end
						resizeButton.limit = 0.01
						
						local x, y = resizeButton:GetRight(), resizeButton:GetBottom()
						local newX, newY = x - resizeButton.x1, resizeButton.y2 - y
						
						STATRACK.Resize(min(TUBE_RACK_MAX_Y, max(TUBE_RACK_MIN_X, x - resizeButton.x1)), min(TUBE_RACK_MAX_Y, max(TUBE_RACK_MIN_Y, resizeButton.y2 - y)))
						
						if abs(newX) > 1.35*TUBE_RACK_MAX_X or abs(newY) > 1.35*TUBE_RACK_MAX_Y or -- The player is probably shaking the mouse sporadically due to a resizing bug...
								resizeButton:GetButtonState() ~= "PUSHED" then 
							resizeButton:SetScript("OnUpdate", nil)
							StopHighlight()
						end
					end
				)
				resizeButton:SetScript("OnMouseUp", StopHighlight)
			elseif button == "LeftButton" and not down then
				StopHighlight()
			end
		end
	)
	resizeButton:Hide()
	
	if hideButton == nil then 
		hideButton = CreateFrame("BUTTON", "STATRACK_HIDE_BUTTON", frame)
	end
	hideButton:SetFrameStrata("HIGH")
	hideButton:SetWidth(16)
	hideButton:SetHeight(16)
	hideButton:SetPoint("TOPRIGHT", frame, "TOPLEFT", frame:GetWidth()*0.02, -frame:GetHeight()*0.1)
	hideButton:SetNormalTexture("Interface/AddOns/StatRack/Textures/hide_button.blp")
	hideButton:SetAlpha(0.5)
	hideButton:RegisterForClicks("LeftButtonUp", "LeftButtonDown")
	hideButton:EnableMouse(true)
	hideButton:SetScript("OnClick", 
		function(self, button, down)
			if button == "LeftButton" and down then
				local StopHighlight = function()
					hideButton:SetAlpha(0.5)
					hideButton:SetScript("OnLeave", nil)
					hideButton:SetScript("OnMouseUp", nil)
				end
				
				hideButton:SetAlpha(0.9)
				hideButton:SetScript("OnLeave", StopHighlight)
				hideButton:SetScript("OnMouseUp", StopHighlight)
			elseif button == "LeftButton" and not down then
				STATRACK.Hide()
			end
		end
	)
	hideButton:Hide()

	if lockButton == nil then 
		lockButton = CreateFrame("BUTTON", "STATRACK_LOCK_BUTTON", frame)
	end
	lockButton:SetFrameStrata("HIGH")
	lockButton:SetWidth(16)
	lockButton:SetHeight(16)
	lockButton:SetPoint("TOPRIGHT", frame, "TOPLEFT", frame:GetWidth()*0.02, -18-frame:GetHeight()*0.1)
	lockButton:SetNormalTexture("Interface/AddOns/StatRack/Textures/locked_button.blp")
	lockButton.locked = true
	lockButton:SetAlpha(0.5)
	lockButton:RegisterForClicks("LeftButtonUp", "LeftButtonDown")
	lockButton:EnableMouse(true)
	lockButton:SetScript("OnClick", 
		function(self, button, down)
			if button == "LeftButton" and down then
				local StopHighlight = function()
					lockButton:SetAlpha(0.5)
					lockButton:SetScript("OnLeave", nil)
					lockButton:SetScript("OnMouseUp", nil)
				end
				
				lockButton:SetAlpha(0.9)
				lockButton:SetScript("OnLeave", StopHighlight)
				lockButton:SetScript("OnMouseUp", StopHighlight)
			elseif button == "LeftButton" and not down then
				if lockButton.locked == true then 
					lockButton.locked = false
					lockButton:SetNormalTexture("Interface/AddOns/StatRack/Textures/unlocked_button.blp")
					frame:RegisterForDrag("LeftButton")
					frame:EnableMouse(true)
					resizeButton:Show()
				else
					lockButton.locked = true
					lockButton:SetNormalTexture("Interface/AddOns/StatRack/Textures/locked_button.blp")
					frame:RegisterForDrag(nil)
					frame:EnableMouse(false)
					resizeButton:Hide()
				end
			end
		end
	)
	lockButton:Hide()


	-- Moving and resizing
	frame:SetMovable(true)
	frame:EnableMouse(true)
	frame:SetScript("OnDragStart", StartMoving)
	frame:SetScript("OnDragStop", StopMovingOrSizing)
end

function STATRACK.Reload()
	isLoaded = false
	STATRACK.Hide() -- TODO: Bubble texture fix for reloading
	STATRACK.SetVar("visible", true, true)
	frame:GetScript("OnEvent")(frame, "PLAYER_ENTERING_WORLD") -- Shortcut to show welcome if resetting addon, informing the player of the reset
	if welcome ~= nil and welcome.animGroup ~= nil and welcome.animGroup.fade ~= nil then 
		welcome.animGroup.fade:SetStartDelay(2)
		welcome.animGroup:Play()
	end
end

function STATRACK.Show()
	if tubeRack == nil then 
		STATRACK.Load()
	end
	
	frame:Show()
	tubeRack:Show()
	for q=1,NUM_TUBE_FRAMES,1 do
		statsDisplay[q][SD_TUBE_FRAME]:Show()
		for r=1,NUM_BUBBLE_TEX,1 do
			statsDisplay[q][SD_TUBE_FRAME].bubbleTextures[r]:Show()
			statsDisplay[q][SD_TUBE_FRAME].letterFrame:Show()
		end
	end
	
	frame:SetScript("OnEnter", 
		function()
			hideButton:Show()
			lockButton:Show()
			if not lockButton.locked then
				resizeButton:ClearAllPoints()
				resizeButton:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT")
				resizeButton:Show()
			else
				frame:EnableMouse(false) -- Allow click throughs after showing the utility menu
			end
			hideButton:SetScript("OnUpdate",
				function()
					if not MouseIsOver(frame, 10, -10, -16, 10) and resizeButton:GetButtonState() ~= "PUSHED" then
						hideButton:Hide()
						lockButton:Hide()
						
						resizeButton:SetAlpha(0.3)
						hideButton:SetScript("OnUpdate", nil)
						frame:EnableMouse(true)
					end
				end
			)
		end
	)
	
	STATRACK.Resize(frame:GetWidth(), frame:GetHeight()) -- TODO: Hotfix for slow-loaded user layout cache. Could also be the best sol'n
	
	frame:SetScript("OnUpdate", StatRackFrameTicker)
	STATRACK.SetVar("visible", true, true)
end

function STATRACK.Hide()
	frame:SetScript("OnUpdate", nil)
	
	frame:SetScript("OnEnter", nil)
	if tubeRack ~= nil then 
		hideButton:SetScript("OnUpdate", nil)
		
		hideButton:Hide()
		lockButton:Hide()
		resizeButton:Hide()
		
		for q=1,NUM_TUBE_FRAMES,1 do
			for r=1,NUM_BUBBLE_TEX,1 do
				statsDisplay[q][SD_TUBE_FRAME].bubbleTextures[r]:Hide()
			end
			statsDisplay[q][SD_TUBE_FRAME]:Hide()
			statsDisplay[q][SD_TUBE_FRAME].letterFrame:Hide()
		end
		tubeRack:Hide()
	end
	frame:Hide()

	frame:SetScript("OnUpdate", nil)
	
	STATRACK.SetVar("visible", false, true)
end

function STATRACK.Resize(x, y)
	local origLeft, origBottom, yDiff = frame:GetLeft(), frame:GetBottom(), frame:GetHeight() - y
	local currWidth, currHeight = tubeRack:GetWidth(), tubeRack:GetHeight()
	--if (abs(currWidth-x) < 0.000001 and abs(currHeight-y) < 0.000001) then resizeButton:SetPoint("BOTTOMRIGHT", tubeRack, "BOTTOMRIGHT") return end
	--print("in", currWidth-x, currHeight-y)
	frame:SetWidth(x)
	frame:SetHeight(y)
	tubeRack:SetWidth(x) -- yuck
	tubeRack:SetHeight(y) -- yuck
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", origLeft, (abs(yDiff) > 0.000001 and origBottom+yDiff or origBottom))
	tubeRack:SetAllPoints(frame)
	-- Initialize Tubes
	for q=1,NUM_TUBE_FRAMES,1 do
		local tubeName, tubeColor, tubeXOffset, tubeLetter = unpack(statTubesData[q])
		local tf = statsDisplay[q][SD_TUBE_FRAME]
		
		tf:SetWidth(frame:GetWidth()*tubeRack.tubeWidthScaleX)
		tf:SetHeight(frame:GetHeight()*tubeRack.tubeHeightScaleY)
		tf:ClearAllPoints()
		tf:SetPoint("CENTER", tubeRack, "CENTER", frame:GetWidth()*tubeXOffset, 0)
		tf:SetPoint("BOTTOM", tubeRack, "BOTTOM", 0, frame:GetHeight()*tubeRack.tubeOffsetY)
		
		tf.letterFrame.fontString:SetTextHeight(tf:GetWidth()*1.1906)
		tf.letterFrame.fontString:SetPoint("BOTTOM", tf, "BOTTOM", 0, frame:GetHeight()*0.52)
		tf.letterFrame.fontString:SetWidth(tf:GetWidth()*1.1)
		tf.letterFrame.fontString:SetHeight(tf:GetWidth()*1.1)
	end
end

function SetNewScaleAnim(tubeFrame, newScale)
	newScale = max(0.0001, min(1, newScale))
	tubeFrame.prevScale = tubeFrame.scale
	tubeFrame.scale = newScale
	local relativeScale = newScale / tubeFrame.prevScale
	--print(relativeScale, (tubeFrame:GetHeight()-(relativeScale*tubeFrame:GetHeight()))*0.5)
	tubeFrame.animGroup.scale:SetScale(1, relativeScale)
	tubeFrame.animGroup.translation:SetOffset(0, -(tubeFrame:GetHeight()-(relativeScale*tubeFrame:GetHeight()))*0.5)
	
	tubeFrame.animGroup.scale:SetScript("OnFinished", 
		function(self)
			--print("Setting new height: "..tubeRack:GetHeight()*tubeRack.tubeHeightScaleY*tubeFrame.scale..". And new bottom offset: "..(tubeRack:GetHeight()*tubeRack.tubeOffsetY*newScale*0.5)..".")
			tubeFrame:SetHeight(tubeRack:GetHeight()*tubeRack.tubeHeightScaleY*tubeFrame.scale)
			tubeFrame.animGroup:Stop()
		end
	)

	tubeFrame.animGroup:Play()
end

function FormatTextFrameText()	
	local SDs = statsDisplay

	return string.format(
		"c: |cff%s%d\n|rh: |cff%s%d\n|rm: |cff%s%d\n|rv: |cff%s%d\n|rz: |cff%s%d",
		SDs[1][4], SDs[1][3], 
		SDs[2][4], SDs[2][3], 
		SDs[3][4], SDs[3][3], 
		SDs[4][4], SDs[4][3], 
		SDs[5][4], SDs[5][3]
	)
end

frame:RegisterEvent("PLAYER_ENTERING_WORLD")
frame:RegisterEvent("ADDONS_UNLOADING")
frame:RegisterEvent("LOADING_SCREEN_ENABLED")
frame:RegisterEvent("LOADING_SCREEN_DISABLED")
frame:RegisterEvent("CINEMATIC_START")
frame:RegisterEvent("CINEMATIC_STOP")
frame:SetScript("OnEvent", 
	function(self, event) 
		if event == "PLAYER_ENTERING_WORLD" and not isLoaded and not InCinematic() then
			local mainStatValue = 0
			local mainStatIndex
			local i = 1
			while i <= 4 do
				local thisStatValue = UnitStat("player", i)
				if mainStatValue < thisStatValue then
					mainStatValue = thisStatValue
					mainStatIndex = i
				end
				i = i * 2 -- Half Life i=i*2 coming soon
			end
			STATRACK.statsDisplay.statIndex = mainStatIndex
			
			if STATRACK.GetVar("visible", true) ~= false then
				STATRACK.Load()
				
				STATRACK.Show()
				
				if welcome == nil then 
					welcome = CreateFrame("FRAME", "STATRACK_WELCOME") -- TODO: Requires frame recycling
					welcome.background = CreateFrame("FRAME", "STATRACK_WELCOME_BG")
					welcome.background.texture = welcome.background:CreateTexture("STATRACK_WELCOME_BG_TEX", "ARTWORK")
					welcome.statRackText = welcome:CreateFontString("STATRACK_WELCOME_SR_FS", "OVERLAY")
					welcome.animGroup = welcome:CreateAnimationGroup("STATRACK_WELCOME_ANIM_GROUP")
					welcome.animGroup.fade = welcome.animGroup:CreateAnimation("ALPHA", "STATRACK_WELCOME_FADE")
				end
				welcome.background:SetPoint("BOTTOM", frame, "BOTTOM")
				--welcome.background:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT")
				welcome.background:SetHeight(frame:GetHeight()*0.14)
				welcome.background:SetWidth(frame:GetWidth()*0.8)
				
				welcome.background.texture:SetAllPoints(welcome.background)
				welcome.background.texture:SetColorTexture(0.0, 0.0, 0.0, 0.65)
				
				welcome.statRackText:SetFontObject("GameFontNormal")
				welcome.statRackText:SetFont("Interface/AddOns/StatRack/Fonts/Cascadia.ttf", max(3, frame:GetHeight()*0.07))
				welcome.statRackText:SetText(string.format("StatRack v%s", GetAddOnMetadata("StatRack", "Version")))
				welcome.statRackText:SetTextColor(1.0, 1.0, 1.0, 1.0)
				welcome.statRackText:SetPoint("CENTER", welcome.background, "CENTER")
				welcome.statRackText:Show()
				welcome.background.texture:Show()
				
				welcome.animGroup.fade:SetDuration(3.5)
				welcome.animGroup.fade:SetToAlpha(0.0)
				welcome.animGroup.fade:SetFromAlpha(1.0)
				welcome.animGroup.fade:SetStartDelay(5)
				
				welcome:Show()
				welcome.background:Show() 
				welcome.statRackText:Show()
				
				welcome:RegisterEvent("LOADING_SCREEN_DISABLED")
				welcome:SetScript("OnEvent", function() welcome.animGroup:Play() welcome:UnregisterAllEvents() welcome:SetScript("OnEvent", nil) end)
				welcome.animGroup:SetScript("OnUpdate", function() if InCinematic() then welcome:Hide() welcome.background:Hide() welcome.statRackText:Hide() end welcome.background:SetAlpha(welcome:GetAlpha()) end)
				welcome.animGroup:SetScript("OnFinished", function() welcome:Hide() welcome.background:Hide() welcome.statRackText:Hide() welcome.animGroup:SetScript("OnFinished", nil) welcome.animGroup:SetScript("OnUpdate", nil) end)
			else
				STATRACK.Hide()
			end
			isLoaded = true
		elseif event == "LOADING_SCREEN_ENABLED" then
			frame:SetScript("OnUpdate", nil)
		elseif event == "LOADING_SCREEN_DISABLED" then
			frame:SetScript("OnUpdate", StatRackFrameTicker)
		elseif event == "ADDONS_UNLOADING" then
			if frame:GetLeft() ~= nil and frame:GetBottom() ~= nil and frame:GetWidth() ~= nil and frame:GetHeight() ~= nil then
				STATRACK.SetVar("userPos", {frame:GetLeft(), frame:GetBottom(), frame:GetWidth(), frame:GetHeight()}, true)
			end
		elseif event == "CINEMATIC_START" then
			STATRACK.ShowAfterCinematic = STATRACK.GetVar("visible", true) ~= false
			STATRACK.Hide()
			if welcome ~= nil then
				welcome:Hide()
			end
		elseif event == "CINEMATIC_STOP" then
			if UnitLevel("player") == 1 then
				isLoaded = false
				frame:GetScript("OnEvent")(frame, "PLAYER_ENTERING_WORLD") -- Manual load after cutsceen
			elseif STATRACK.ShowAfterCinematic then
				STATRACK.Show()
			end
		end
	end
)